use std::pin::Pin;
use std::task::{Context, Poll};
use std::sync::{Arc, Mutex, MutexGuard};

use futures::prelude::*;
use futures_signals::signal::{Mutable, Signal, MutableSignalCloned};

use termion::event::Key;

use crate::input::Inputs;
use crate::schema::Api;

use slog::{
    Drain,
    Level,
    Record,
    OwnedKVList,
};

use tui::text::{Span, Spans};
use tui::style::{Style, Color};

#[derive(PartialEq, Eq, Debug, Clone)]
pub enum Window {
    Main,
    Help,
}

enum ConnState<F> {
    Connecting(F),
    Connected(Api)
}

/// Application state struct
pub struct Sute<'a, S, F> {

    // TODO: BE SMART. Inputs change the state, resize signals change the state, futures completing
    // change the state.

    pub state: Mutable<SuteState<'a>>,
    statesig: MutableSignalCloned<SuteState<'a>>,
    signal: S,
    inputs: Inputs,
    api: Option<ConnState<F>>,
    new: bool
}

impl<'a, S: Unpin, F: Unpin> Sute<'a, S, F> {
    pub fn new(s: S, log: Arc<LogDrain<'a>>, api: F) -> Self {
        let inputs = Inputs::new();
        let state = Mutable::new(SuteState::new(log));

        Self { 
            statesig: state.signal_cloned(),
            state: state,
            signal: s,
            inputs: inputs,
            api: Some(ConnState::Connecting(api)),
            new: true,
        }
    }

    fn handle_resize(&mut self, new_size: (u16,u16)) {
        (self.state.lock_mut()).size = new_size;
    }

    fn handle_input(&mut self, key: Key) {
        // TODO modify signal implementation so we don't have to modify the state on all ticks.
        let mut state = self.state.lock_mut();
        state.tick = state.tick + 1;
        match key {
            Key::Char('q') => state.running = false,
            Key::Char('?') => state.active_win = Window::Help,
            Key::Esc => {
                if state.active_win == Window::Help {
                    state.active_win = Window::Main;
                }
            }
            _ => {}
        }
    }

    pub fn get_state(&self) -> SuteState<'a> {
        self.state.get_cloned()
    }
}

impl<'a, S: Signal<Item=(u16,u16)> + Unpin, F: Future<Output=Api> + Unpin> Signal for Sute<'a, S, F> {
    type Item = SuteState<'a>;
    fn poll_change(mut self: Pin<&mut Self>, cx: &mut Context) -> Poll<Option<Self::Item>> {
        if let Poll::Ready(Some(key)) = Pin::new(&mut self.inputs).poll_next(cx) {
            self.handle_input(key);
        }
        if let Poll::Ready(Some(size)) = Pin::new(&mut self.signal).poll_change(cx) {
            self.handle_resize(size);
        }
        if let Some(ConnState::Connecting(mut apif)) = self.api.take() {
            if let Poll::Ready(api) = Pin::new(&mut apif).poll(cx) {
                self.api = Some(ConnState::Connected(api));
            } else {
                self.api = Some(ConnState::Connecting(apif));
            }
        }

        // TODO chunk this?
        Pin::new(&mut self.statesig).poll_change(cx)
    }
}

#[derive(Debug, Clone)]
pub struct SuteState<'a> {
    pub active_win: Window,
    pub size: (u16,u16),
    pub running: bool,
    pub tick: usize,
    pub server: Option<String>,
    pub log: Arc<LogDrain<'a>>,
}

impl<'a> SuteState<'a> {
    pub fn new(log: Arc<LogDrain<'a>>) -> Self {
        Self {
            active_win: Window::Main,
            size: (80,20),
            running: true,
            tick: 0,
            server: None,
            log: log,
        }
    }
}

#[derive(Debug)]
pub struct LogDrain<'a> {
    inner: Mutex<Vec<Spans<'a>>>,
}

impl<'a> LogDrain<'a> {
    pub fn new() -> Self {
        Self { inner: Mutex::new(Vec::new()) }
    }
    pub fn get_inner(&self) -> MutexGuard<Vec<Spans<'a>>> {
        self.inner.lock().unwrap()
    }
}
impl<'a> Drain for LogDrain<'a> {
    type Ok = ();
    type Err = ();
    fn log(&self, record: &Record, values: &OwnedKVList) -> Result<Self::Ok, Self::Err> {
        let critical_style: Style = Style::default().fg(Color::Magenta);
        let error_style: Style = Style::default().fg(Color::Red);
        let warning_style: Style = Style::default().fg(Color::Yellow);
        let info_style: Style = Style::default().fg(Color::Cyan);
        let debug_style: Style = Style::default().fg(Color::White);
        let trace_style: Style = Style::default().fg(Color::Green);

        let lvlspan = match record.level() {
            Level::Critical => Span::styled("CRIT", critical_style),
            Level::Error => Span::styled("ERRO", error_style),
            Level::Warning => Span::styled("WARN", warning_style),
            Level::Info => Span::styled("INFO", info_style),
            Level::Debug => Span::styled("DEBG", debug_style),
            Level::Trace => Span::styled("TRCE", trace_style),
        };
        let contentspan = Span::from(record.msg().to_string());

        {
            let mut v = self.inner.lock().unwrap();
            v.push(Spans::from(vec![
                Span::from("["), lvlspan, Span::from("]: "),
                contentspan,
            ]));
        }

        Ok(())
    }
}

